"""Beamline Control."""

import time as _time

from .device import Device as _Device


class ASPPSCtrl(_Device):
    """Accelerator PPS Control."""

    DEVICE = 'AS-Glob:PP-Summary'

    PROPERTIES_DEFAULT = (
        'TunAccessWaitTime-Cte',
        'TunAccessWaitTimeLeft-Mon',
        'TunAccessWaitTimeElapsed-Mon',
    )

    def __init__(self, props2init='all', **kwargs):
        """Init."""
        super().__init__(ASPPSCtrl.DEVICE, props2init=props2init, **kwargs)

    @property
    def tunnel_access_wait_time(self):
        """Return total wait time for tunnel access [s]."""
        return self['TunAccessWaitTime-Cte']

    @property
    def tunnel_access_wait_time_left(self):
        """Return wait time left for tunnel access [s]."""
        return self['TunAccessWaitTimeLeft-Mon']

    @property
    def tunnel_access_wait_time_elapsed(self):
        """Return wait time elapsed for tunnel access [s]."""
        return self['TunAccessWaitTimeElapsed-Mon']


class ASMPSCtrl(_Device):
    """Accelerator MPS Control."""

    DEVICE = 'AS-Glob:MP-Summary'

    PROPERTIES_DEFAULT = (
        'AlarmGammaShutter-Mon',
        'DsblGamma-Cmd', 'EnblGamma-Cmd',
        # 'DsblGamma-Mon', 'EnblGamma-Mon',
        )

    def __init__(self, props2init='all', **kwargs):
        """Init."""
        super().__init__(ASMPSCtrl.DEVICE, props2init=props2init, **kwargs)

    @property
    def gamma_enabled(self):
        """."""
        return self['AlarmGammaShutter-Mon'] == 1

    @property
    def gamma_disabled(self):
        """."""
        return self['AlarmGammaShutter-Mon'] == 0

    def cmd_gamma_enable(self, timeout=None):
        """Enable gamma signal for beamlines."""
        self['EnblGamma-Cmd'] = 1
        return self._wait('AlarmGammaShutter-Mon', 1, timeout)

    def cmd_gamma_disable(self, timeout=None):
        """Disable gamma signal for beamlines."""
        self['DsblGamma-Cmd'] = 1
        return self._wait('AlarmGammaShutter-Mon', 0, timeout)


class BLInterlockCtrl(_Device):
    """Beamline Interlock Control."""

    TIMEOUT_GATEVALVE = 20  # [s]
    TIMEOUT_SHUTTER = 7  # [s]
    TIMEOUT_EPS_RESET = 3  # [s]

    _DEF_MSG_SUCCESS = 'Ok'

    class DEVICES:
        """Devices names."""

        CAX = 'CAX'
        ALL = (CAX, )

    PROPERTIES_DEFAULT = (
        # Status da liberação do gamma pela máquina – 1 indica liberado
        'M:PPS01:HABILITACAO_MAQUINA',

        # general EPS status for frontend shutter
        'F:PPS01:HABILITACAO_EPS_GS_X',

        # general EPS status for hutchA shutter
        'A:PPS01:HABILITACAO_EPS',

        # EPS status
        'A:EPS01:StatusPos',
        'A:EPS01:StatusTemp',
        'A:EPS01:StatusVac',
        'B:EPS01:StatusVac',
        'F:EPS01:StatusPos',
        'F:EPS01:StatusTemp',
        'F:EPS01:StatusVac',

        # open front-end and first hutchA valve gates
        'F:EPS01:openGates',

        'A:EPS01:GV5open', 'A:EPS01:GV5closed',  # open/close gate status
        'F:EPS01:GV4open', 'F:EPS01:GV4closed',  # open/close gate status
        'F:EPS01:GV3open', 'F:EPS01:GV3closed',  # open/close gate status
        # 'F:EPS01:GV2open', 'F:EPS01:GV2closed',  # open/close gate status
        'F:EPS01:GV1open', 'F:EPS01:GV1closed',  # open/close gate status

        # reset EPS
        'F:EPS01:resetCmd',

        # Enviar 1 para abrir o Front-end completo (abre o gamma e o photon)
        'F:PPS01:FOEOPEN',

        # Enviar 1 para fechar o Front-end completo (fecha o photon e o gamma)
        'F:PPS01:FOECLOSE',

        # Status photon shutter front-end – 0 indica aberto
        'F:PPS01:PS_STATUS',

        # Status gamma shutter front-end – 0 indica aberto
        'F:PPS01:GS_X_STATUS',

        # shutter da cabana a
        # - Enviar 1 – se estiver fechado abre – se estiver aberto fecha
        'A:PPS01:OEAOPENCLOSE',
        # Status do gamma shutter da cabana A – 0 indica aberto
        'A:PPS01:PG_STATUS',
        # status da procura cabana A – 1 indica procura feita
        'A:PPS01:SEARCH_OK',
        # status da procura cabana B – 1 indica procura feita
        'B:PPS01:SEARCH_OK',
        # open other hutchA and hutchB valve gates
        'B:EPS01:openGates',
        'B:EPS01:GV7open', 'B:EPS01:GV7closed',  # open/close gate status
        'A:EPS01:GV6open', 'A:EPS01:GV6closed',  # open/close gate status
        )

    def __init__(self, devname=None, props2init='all', **kwargs):
        """Init."""
        if devname is None:
            devname = self.DEVICES.CAX
        if devname not in self.DEVICES.ALL:
            raise NotImplementedError(devname)
        self._error_log = self._DEF_MSG_SUCCESS
        super().__init__(devname, props2init=props2init, **kwargs)

    @property
    def error_log(self):
        """Return error log for last check method invoked."""
        return self._error_log

    @property
    def is_hutchA_intlk_search_done(self):
        """."""
        msg = 'hutch A interlock search not Ok.'
        return self._check_set_error_log(self['A:PPS01:SEARCH_OK'] == 1, msg)

    @property
    def is_hutchB_intlk_search_done(self):
        """."""
        msg = 'hutch B interlock search not Ok.'
        return self._check_set_error_log(self['B:PPS01:SEARCH_OK'] == 1, msg)

    @property
    def is_machine_gamma_enabled(self):
        """."""
        msg = 'machine gamma not enabled.'
        return self._check_set_error_log(
            self['M:PPS01:HABILITACAO_MAQUINA'] == 1, msg)

    @property
    def is_frontend_gamma_shutter_opened(self):
        """."""
        msg = 'front-end gamma shutter not opened.'
        return self._check_set_error_log(self['F:PPS01:GS_X_STATUS'] == 0, msg)

    @property
    def is_frontend_photon_shutter_opened(self):
        """."""
        msg = 'front-end photon shutter not opened.'
        return self._check_set_error_log(self['F:PPS01:PS_STATUS'] == 0, msg)

    @property
    def is_hutchA_gamma_shutter_opened(self):
        """."""
        msg = 'hutch A gamma shutter not opened.'
        return self._check_set_error_log(self['A:PPS01:PG_STATUS'] == 0, msg)

    @property
    def is_hutchA_eps_dvf_pos_ok(self):
        """."""
        msg = 'hutch A EPS DVF position not Ok.'
        return self._check_set_error_log(bool(self['A:EPS01:StatusPos']), msg)

    @property
    def is_hutchA_eps_temperatures_ok(self):
        """."""
        msg = 'hutch A EPS temperatures not Ok.'
        return self._check_set_error_log(bool(self['A:EPS01:StatusTemp']), msg)

    @property
    def is_hutchA_eps_vacuum_ok(self):
        """."""
        msg = 'hutch A EPS vacuum not Ok.'
        return self._check_set_error_log(bool(self['A:EPS01:StatusVac']), msg)

    @property
    def is_hutchB_eps_vacuum_ok(self):
        """."""
        msg = 'hutch B EPS vacuum not Ok.'
        return self._check_set_error_log(bool(self['B:EPS01:StatusVac']), msg)

    @property
    def is_frontend_eps_mirror_pos_ok(self):
        """."""
        msg = 'front-end EPS mirror position not Ok.'
        return self._check_set_error_log(bool(self['F:EPS01:StatusPos']), msg)

    @property
    def is_frontend_eps_temperatures_ok(self):
        """."""
        msg = 'front-end EPS temperatures not Ok.'
        return self._check_set_error_log(bool(self['F:EPS01:StatusTemp']), msg)

    @property
    def is_frontend_eps_vacuum_ok(self):
        """."""
        msg = 'front-end EPS vacuum not Ok.'
        return self._check_set_error_log(bool(self['F:EPS01:StatusVac']), msg)

    @property
    def is_frontend_gatevalves_opened(self):
        """."""
        msg = 'front-end gatevalve A:EPS01:GV5 not opened.'
        if not self._check_set_error_log(bool(self['A:EPS01:GV5open']), msg):
            return False
        msg = 'front-end gatevalve F:EPS01:GV4 not opened.'
        if not self._check_set_error_log(bool(self['F:EPS01:GV4open']), msg):
            return False
        msg = 'front-end gatevalve F:EPS01:GV3 not opened.'
        if not self._check_set_error_log(bool(self['F:EPS01:GV3open']), msg):
            return False
        # NOTE: GV2open not installed yet.
        # msg = 'front-end gatevalve F:EPS01:GV2 not opened.'
        # if not self._check_set_error_log(bool(self['F:EPS01:GV2open']), msg):
        #     return False
        msg = 'front-end gatevalve F:EPS01:GV1 not opened.'
        if not self._check_set_error_log(bool(self['F:EPS01:GV1open']), msg):
            return False
        return True

    @property
    def is_frontend_gatevalves_closed(self):
        """."""
        if not bool(self['A:EPS01:GV5closed']):
            return False
        if not bool(self['F:EPS01:GV4closed']):
            return False
        if not bool(self['F:EPS01:GV3closed']):
            return False
        # if not bool(self['F:EPS01:GV2open']):
        #     return False
        if not bool(self['F:EPS01:GV1closed']):
            return False
        return True

    @property
    def is_hutchB_gatevalves_opened(self):
        """."""
        msg = 'hutch B gatevalve B:EPS01:GV7 not opened.'
        if not self._check_set_error_log(bool(self['B:EPS01:GV7open']), msg):
            return False
        msg = 'hutch B gatevalve B:EPS01:GV6 not opened.'
        if not self._check_set_error_log(bool(self['A:EPS01:GV6open']), msg):
            return False
        return True

    @property
    def is_hutchB_gatevalves_closed(self):
        """."""
        if not bool(self['B:EPS01:GV7closed']):
            return False
        if not bool(self['A:EPS01:GV6closed']):
            return False
        return True

    @property
    def is_frontend_shutter_eps_permission_ok(self):
        """."""
        msg = 'front-end shutter EPS permission not Ok.'
        return self._check_set_error_log(
            bool(self['F:PPS01:HABILITACAO_EPS_GS_X']), msg)

    @property
    def is_hutchA_shutter_eps_permission_ok(self):
        """."""
        msg = 'hutch A EPS permission not Ok.'
        return self._check_set_error_log(
            bool(self['A:PPS01:HABILITACAO_EPS']), msg)

    @property
    def is_frontend_eps_ok(self):
        """."""
        if self.is_frontend_eps_mirror_pos_ok and \
                self.is_frontend_eps_mirror_pos_ok and \
                self.is_frontend_eps_temperatures_ok and \
                self.is_frontend_eps_vacuum_ok:
            return True
        else:
            return False

    @property
    def is_hutchA_eps_ok(self):
        """."""
        if self.is_hutchA_eps_dvf_pos_ok and \
                self.is_hutchA_eps_temperatures_ok and \
                self.is_hutchA_eps_vacuum_ok:
            return True
        else:
            return False

    @property
    def is_hutchB_eps_ok(self):
        """."""
        return self.is_hutchB_eps_vacuum_ok

    @property
    def is_beamline_eps_ok(self):
        """."""
        if self.is_frontend_eps_ok and \
                self.is_hutchA_eps_ok and \
                self.is_hutchB_eps_ok:
            return True
        else:
            return False

    @property
    def is_beamline_opened(self):
        """Return whether BL is opened."""
        return self.is_hutchA_intlk_search_done and \
            self.is_hutchB_intlk_search_done and \
            self.is_machine_gamma_enabled and \
            self.is_beamline_eps_ok and \
            self.is_frontend_shutter_eps_permission_ok and \
            self.is_hutchA_shutter_eps_permission_ok and \
            self.is_frontend_gamma_shutter_opened and \
            self.is_frontend_photon_shutter_opened and \
            self.is_hutchA_gamma_shutter_opened

    def cmd_beamline_eps_reset(self):
        """."""
        self['F:EPS01:resetCmd'] = 1

    def cmd_frontend_gatevalves_open(self, timeout=None):
        """."""
        if self.is_frontend_gatevalves_opened:
            return True
        else:
            self['F:EPS01:openGates'] = 1

        if timeout is None:
            return True
        t0 = _time.time()
        while not self.is_frontend_gatevalves_opened:
            if _time.time() - t0 > timeout:
                return False
            _time.sleep(0.5)

        return True

    def cmd_hutchB_gatevalves_open(self, timeout=None):
        """."""
        if self.is_hutchB_gatevalves_opened:
            return True
        else:
            self['B:EPS01:openGates'] = 1

        if timeout is None:
            return True
        t0 = _time.time()
        while not self.is_hutchB_gatevalves_opened:
            if _time.time() - t0 > timeout:
                print(self.error_log)
                return False
            _time.sleep(0.5)

        return True

    def cmd_frontend_gamma_and_photon_open(self, timeout=None):
        """."""
        if self.is_frontend_gamma_shutter_opened and \
                self.is_frontend_photon_shutter_opened:
            return True
        else:
            self['F:PPS01:FOEOPEN'] = 1

        if timeout is None:
            return True
        t0 = _time.time()
        while \
                not self.is_frontend_gamma_shutter_opened or \
                not self.is_frontend_photon_shutter_opened:
            if _time.time() - t0 > timeout:
                print(self.error_log)
                return False
            _time.sleep(0.5)

        return True

    def cmd_frontend_gamma_and_photon_close(self, timeout=None):
        """."""
        if \
                not self.is_frontend_gamma_shutter_opened and \
                not self.is_frontend_photon_shutter_opened:
            return True
        else:
            self['F:PPS01:FOECLOSE'] = 1

        if timeout is None:
            return True
        t0 = _time.time()
        while \
                self.is_frontend_gamma_shutter_opened or \
                self.is_frontend_photon_shutter_opened:
            if _time.time() - t0 > timeout:
                print('close frontend shutter timeout reached!')
                return False
            _time.sleep(0.5)

        return True

    def cmd_hutchA_photon_open(self, timeout=None):
        """."""
        if self.is_hutchA_gamma_shutter_opened:
            return True
        else:
            self['A:PPS01:OEAOPENCLOSE'] = 1

        if timeout is None:
            return True
        t0 = _time.time()
        while not self.is_hutchA_gamma_shutter_opened:
            if _time.time() - t0 > timeout:
                print(self.error_log)
                return False
            _time.sleep(0.5)

        return True

    def cmd_hutchA_photon_close(self, timeout=None):
        """."""
        if not self.is_hutchA_gamma_shutter_opened:
            return True
        else:
            self['A:PPS01:OEAOPENCLOSE'] = 1

        if timeout is None:
            return True
        t0 = _time.time()
        while self.is_hutchA_gamma_shutter_opened:
            if _time.time() - t0 > timeout:
                print('close hutchA photon shutter timeout reached!')
                return False
            _time.sleep(0.5)

        return True

    def cmd_beamline_open(self):
        """."""
        if not self.is_hutchA_intlk_search_done:
            print(self.error_log)
            return False

        if not self.is_hutchB_intlk_search_done:
            print(self.error_log)
            return False

        if not self.is_machine_gamma_enabled:
            print(self.error_log)
            return False

        # check and reset EPS
        if not self.is_beamline_eps_ok:
            self.cmd_beamline_eps_reset()
            t0 = _time.time()
            while not self.is_beamline_eps_ok:
                if _time.time() - t0 > self.TIMEOUT_EPS_RESET:
                    print(self.error_log)
                    return False
                _time.sleep(0.5)

        # check frontend shutter permission and open gatevalves for hutchA
        if not self.is_frontend_shutter_eps_permission_ok:
            # open frontend and hutchA gatevalves
            self.cmd_frontend_gatevalves_open()
            t0 = _time.time()
            while not self.is_frontend_gatevalves_opened:
                if _time.time() - t0 > self.TIMEOUT_GATEVALVE:
                    print(self.error_log)
                    return False
                _time.sleep(0.5)

        # check hutchA shutter permission and open gatevalves for hutchB
        if not self.is_hutchA_shutter_eps_permission_ok:
            is_ok = self.cmd_hutchB_gatevalves_open(
                timeout=self.TIMEOUT_GATEVALVE)
            if not is_ok:
                print(self.error_log)
                return False

        # open frontend gamma and photon shutter
        print('open frontend gamma and photon shutters')
        is_ok = self.cmd_frontend_gamma_and_photon_open(
            timeout=self.TIMEOUT_SHUTTER)
        if not is_ok:
            print(self.error_log)
            return False

        # open hutchA photon shutter
        print('open hutchA photon shutter')
        is_ok = self.cmd_hutchA_photon_open(timeout=self.TIMEOUT_SHUTTER)
        if not is_ok:
            print(self.error_log)
            return False

        return True

    def _set_error_log(self, msg):
        # NOTE: should we prefix the msg with timestamp?
        self._error_log = msg

    def _check_set_error_log(self, cond, msg_fail, msg_succeed=None):
        msg_succeed = BLInterlockCtrl._DEF_MSG_SUCCESS if msg_succeed is None \
            else msg_succeed
        if not cond:
            self._set_error_log(msg_fail)
            return False
        else:
            self._set_error_log(msg_succeed)
            return True



"""Implementation of web server data retrieval functions."""
import re as _re
import urllib.request as _urllib_request

from .. import envars as _envars


_TIMEOUT = 5.0  # [seconds]
_EXCDAT_FOLDER = '/magnet/excitation-data/'
_MAGNET_FOLDER = '/magnet/'
_PS_FOLDER = '/pwrsupply/'
_BBB_FOLDER = '/beaglebone/'
_PSTYPES_DATA_FOLDER = '/pwrsupply/pstypes-data/'
_DIAG_FOLDER = '/diagnostics/'
_TIMESYS_FOLDER = '/timesys/'
_MAC_SCHEDULE_FOLDER = '/macschedule/'
_DOC_SERV_FOLDER = '/documentation/services/'


def read_url(url, timeout=_TIMEOUT):
    """Read URL from server."""

    # build list with servers
    urls = [_envars.SRVURL_CSCONSTS + url, _envars.SRVURL_CSCONSTS_2 + url]
    connected = False
    for url_ in urls:
        try:
            # try a new server
            response = _urllib_request.urlopen(url_, timeout=timeout)
            data = response.read()
            text = data.decode('utf-8')
            connected = True
            break
        except Exception:
            # could not connect with current server
            print('Error reading url "' + url_ + '"!')
    if not connected:
        raise Exception('Error reading web servers!')
    return text


def server_online():
    """Verify if the server is online."""
    try:
        read_url('', timeout=_TIMEOUT)
        return True
    except Exception:
        return False


def magnets_model_data(timeout=_TIMEOUT):
    """Return the text of the retrieved magnet model data."""
    url = _MAGNET_FOLDER + 'magnets-model-data.txt'
    return read_url(url, timeout=timeout)


def magnets_excitation_data_read(filename, timeout=_TIMEOUT):
    """Return the text of the retrieved magnet excitation data."""
    return read_url(_EXCDAT_FOLDER + filename, timeout=timeout)


def magnets_excitation_ps_read(timeout=_TIMEOUT):
    """Return the power supply excitation data."""
    url = _MAGNET_FOLDER + 'magnet-excitation-ps.txt'
    return read_url(url, timeout=timeout)


def ps_pstypes_names_read(timeout=_TIMEOUT):
    """Return the text of the power supplies type."""
    url = _PS_FOLDER + 'pstypes-names.txt'
    return read_url(url, timeout=timeout)


def ps_pstype_data_read(filename, timeout=_TIMEOUT):
    """Return the power supply data."""
    url = _PSTYPES_DATA_FOLDER + filename
    return read_url(url, timeout=timeout)


def ps_pstype_setpoint_limits(timeout=_TIMEOUT):
    """Return the power supply setpoint limits data."""
    url = _PS_FOLDER + 'pstypes-setpoint-limits.txt'
    return read_url(url, timeout=timeout)


def pu_pstype_setpoint_limits(timeout=_TIMEOUT):
    """Return the power supply setpoint limits data."""
    url = _PS_FOLDER + 'putypes-setpoint-limits.txt'
    return read_url(url, timeout=timeout)


def ps_psmodels_read(timeout=_TIMEOUT):
    """Return the psmodels file."""
    url = _PS_FOLDER + 'psmodels.txt'
    return read_url(url, timeout=timeout)


def ps_siggen_configuration_read(timeout=_TIMEOUT):
    """Return power supplies signal default generation configuration."""
    url = _PS_FOLDER + 'siggen-configuration.txt'
    return read_url(url, timeout=timeout)


def pu_psmodels_read(timeout=_TIMEOUT):
    """Return the pumodels file."""
    url = _PS_FOLDER + 'pumodels.txt'
    return read_url(url, timeout=timeout)


def beaglebone_freq_mapping(timeout=_TIMEOUT):
    """Return the beaglebone Black BSMP PRU sync off and on freqs."""
    url = _BBB_FOLDER + 'beaglebone-freq.txt'
    return read_url(url, timeout=timeout)


def beaglebone_ip_list(timeout=_TIMEOUT):
    """Return the beaglebone Black IP list."""
    url = _BBB_FOLDER + 'ip-list.txt'
    return read_url(url, timeout=timeout)


def bbb_udc_mapping(timeout=_TIMEOUT):
    """Read beaglebone-udc mapping."""
    url = _BBB_FOLDER + 'beaglebone-udc.txt'
    return read_url(url, timeout=timeout)


def udc_ps_mapping(timeout=_TIMEOUT):
    """Read beaglebone-udc mapping."""
    url = _BBB_FOLDER + 'udc-bsmp.txt'
    return read_url(url, timeout=timeout)


def crates_mapping(timeout=_TIMEOUT):
    """Return the crates mapping."""
    url = _DIAG_FOLDER + 'microTCA-vs-BPMs-mapping/'
    text = read_url(url, timeout=timeout)
    pat = _re.compile('"(names_crate[a-zA-Z_0-9]*.cfg)"')
    files = pat.findall(text)
    txt = ''
    for fi in files:
        for time in read_url(url + fi, timeout=timeout).splitlines():
            txt += f'{time:<40s}' + f'{fi[6:13]:>10s}\n'
        txt += '\n\n'
    return txt


def bpms_data(timeout=_TIMEOUT):
    """Return the BPMs data."""
    url = _DIAG_FOLDER + 'bpms-data.txt'
    return read_url(url, timeout=timeout)


def timing_devices_mapping(timeout=_TIMEOUT):
    """Return the timing devices connections mapping."""
    url = _TIMESYS_FOLDER + 'timing-devices-connection.txt'
    return read_url(url, timeout=timeout)


def high_level_triggers(timeout=_TIMEOUT):
    """Return the data defining the high level triggers."""
    url = _TIMESYS_FOLDER + 'high-level-triggers.py'
    return read_url(url, timeout=timeout)


def high_level_events(timeout=_TIMEOUT):
    """Return the data defining the high level events."""
    url = _TIMESYS_FOLDER + 'high-level-events.py'
    return read_url(url, timeout=timeout)


def bsmp_dclink_mapping(timeout=_TIMEOUT):
    """Read bsmp dclink mapping."""
    url = _PS_FOLDER + 'bsmp-dclink.txt'
    return read_url(url, timeout=timeout)


def mac_schedule_read(year, timeout=_TIMEOUT):
    """Read machine schedule data."""
    url = _MAC_SCHEDULE_FOLDER + str(year) + '.txt'
    return read_url(url, timeout=timeout)


def doc_services_read(timeout=_TIMEOUT):
    """Read service documentation data."""
    url = _DOC_SERV_FOLDER
    text = read_url(url, timeout=timeout)
    pattern = _re.compile('"([a-zA-Z_0-9]*.yml)"')
    files = pattern.findall(text)
    text = ''
    for file in files:
        text += read_url(url + file, timeout=timeout)
        text += '\n\n'
    return text



"""Devices subpackage."""

from .afc_acq_core import AFCPhysicalTrigger, AFCACQLogicalTrigger
from .bbb import BunchbyBunch
from .bpm import BPM
from .bpm_fam import FamBPMs
from .bpm_eq import EqualizeBPMs
from .currinfo import CurrInfoTranspEff, CurrInfoLinear, CurrInfoBO, \
    CurrInfoSI, CurrInfoAS
from .dcct import DCCT
from .device import Device, DeviceSet
from .egun import EGBias, EGFilament, EGHVPS, EGTriggerPS, EGPulsePS, EGun
from .energy import Energy
from .fofb import FOFBCtrlDCC, BPMDCC, FOFBCtrlRef, FamFOFBControllers, \
    FamFastCorrs, HLFOFB
from .fofb_acq import FOFBCtrlSysId, FOFBPSSysId, FamFOFBSysId, \
    FOFBCtrlLamp, FOFBPSLamp, FamFOFBLamp
from .ict import ICT, TranspEff
from .ids import IDBase, APU, WIG, PAPU, EPU, DELTA, ID
from .idff import IDFF
from .injctrl import InjCtrl
from .injsys import PUMagsStandbyHandler, BOPSRampStandbyHandler, \
    BORFRampStandbyHandler, InjSysStandbyHandler, LinacStandbyHandler, \
    InjSysPUModeHandler
from .lillrf import LILLRF, DevLILLRF
from .machshift import MachShift
from .modltr import LIModltr
from .orbit_interlock import BPMOrbitIntlk, BaseOrbitIntlk, OrbitInterlock
from .posang import PosAng
from .psconv import PSProperty, StrengthConv
from .pssofb import PSCorrSOFB, PSApplySOFB
from .pwrsupply import PowerSupply, PowerSupplyPU, PowerSupplyFC, \
    PowerSupplyFBP
from .rf import RFGen, ASLLRF, BORFCavMonitor, SIRFCavMonitor, RFCav, \
    RFKillBeam, SILLRFPreAmp, BOLLRFPreAmp, SIRFDCAmp, BORFDCAmp, \
    SIRFACAmp, BORF300VDCAmp
from .screen import Screen
from .sofb import SOFB
from .syncd import DevicesSync
from .timing import EVG, Event, Trigger, HLTiming
from .tune import TuneFrac, TuneProc, Tune, TuneCorr
from .dvf import DVF, DVFImgProc
from .lienergy import LIEnergy
from .intlkctrl import ASPPSCtrl, ASMPSCtrl, BLInterlockCtrl
from .scraper import ScraperH, ScraperV
from .blm import BLM


del device, bpm, dcct, egun, ict, lillrf, modltr
del pwrsupply, posang, psconv, pssofb, rf, injsys, injctrl
del screen, tune, sofb, timing, syncd, energy
del ids, currinfo, bbb, machshift, dvf, lienergy, intlkctrl, scraper
del blm



#!/usr/bin/env python-sirius

"""Unittest module for callbacks.py."""

from unittest import mock, TestCase
import siriuspy.util as util
import siriuspy.callbacks as callbacks


public_interface = (
    'Callback',
)


class TestCallbacks(TestCase):
    """Test util."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
                                        callbacks, public_interface)
        self.assertTrue(valid)


class TestCallbackClass(TestCase):
    """Test Callback Class."""

    public_interface = (
        'add_callback',
        'has_callbacks',
        'remove_callback',
        'clear_callbacks',
        'run_callbacks',
        'run_callback',
        )

    def test_public_interface(self):
        """Test class's public interface."""
        valid = util.check_public_interface_namespace(
            callbacks.Callback, TestCallbackClass.public_interface)
        self.assertTrue(valid)

    def test_constructor(self):
        """Test class constructor."""
        self.assertRaises(ValueError, callbacks.Callback, 2)
        self.assertRaises(ValueError, callbacks.Callback, [2, 4])
        c = callbacks.Callback(self._run_tes0)
        self.assertEqual(1, len(c._callbacks))
        c = callbacks.Callback([self._run_tes0, self._run_tes1])
        self.assertEqual(2, len(c._callbacks))

    def test_add_callback(self):
        """Test add_callback method."""
        c = callbacks.Callback()
        self.assertRaises(ValueError, c.add_callback, callback=4)
        self.assertEqual(0, c.add_callback(callback=lambda x: x))
        self.assertEqual(
            4, c.add_callback(callback=lambda x: x, index=4))
        self.assertEqual(
            5, c.add_callback(callback=lambda x: x, a=4))

    def test_remove_callback(self):
        """Test remove_callback method."""
        c = callbacks.Callback()
        ind = c.add_callback(callback=lambda x: x)
        ind = c.add_callback(callback=lambda x: x)
        c.remove_callback(index=ind)
        self.assertEqual(1, len(c._callbacks))
        c.remove_callback()
        self.assertEqual(1, len(c._callbacks))
        c.remove_callback(index=ind)
        self.assertEqual(1, len(c._callbacks))

    def test_clear_callbacks(self):
        """Test clear_callbacks method."""
        c = callbacks.Callback()
        c.add_callback(callback=lambda x: x)
        c.add_callback(callback=lambda x: x)
        c.clear_callbacks()
        self.assertEqual(0, len(c._callbacks))

    def test_run_callbacks(self):
        """Test run_callbacks method."""
        self.TES = 0
        c = callbacks.Callback()
        c.add_callback(callback=self._run_tes0)
        c.run_callbacks(2)
        self.assertEqual(self.TES, 2)
        c.add_callback(callback=self._run_tes0)
        c.run_callbacks(3)
        self.assertEqual(self.TES, 8)

    def test_run_callback(self):
        """Test run_callback method."""
        self.TES = 0
        c = callbacks.Callback()
        ind1 = c.add_callback(callback=self._run_tes1)
        ind2 = c.add_callback(callback=self._run_tes2)
        c.run_callback(ind1)
        self.assertEqual(self.TES, 1)
        c.run_callbacks(ind2)
        self.assertEqual(self.TES, 2)

    def _run_tes0(self, a, *args, **kwargs):
        self.TES += a

    def _run_tes1(self, *args, **kwargs):
        self.TES = 1

    def _run_tes2(self, *args, **kwargs):
        self.TES = 2



"""MA data class definition."""

from ..search import MASearch as _MASearch
from ..namesys import SiriusPVName as _SiriusPVName
from ..pwrsupply.data import PSData as _PSData


class MAData:
    """MAData Class.

    Class objects that group magnet data.
    """

    def __init__(self, maname):
        """Init method."""
        self._maname = _SiriusPVName(maname)
        psnames = _MASearch.conv_maname_2_psnames(self._maname)
        self._psdata = {}
        for psname in psnames:
            self._psdata[psname] = _PSData(psname=psname)

    @property
    def maname(self):
        """Magnet name."""
        return self._maname

    @property
    def psnames(self):
        """List of power supply names that excite the magnet."""
        return list(self._psdata.keys())

    def magfunc(self, psname):
        """Return magnet function."""
        return self._psdata[psname].magfunc

    def excdata(self, psname):
        """Return excitation data for a power supply."""
        return self._psdata[psname].excdata

    def __getitem__(self, psname):
        """Get item."""
        return self._psdata[psname]



"""."""
import logging as _log

import GPy as gpy
import numpy as _np
from epics.ca import CAThread as _Thread
from numpy.polynomial import polynomial as _np_poly

from .csdev import Const as _Const, get_biasfb_database as _get_database


class BiasFeedback:
    """."""

    def __init__(self, injctrl):
        """."""
        db_ = _get_database()
        self.database = db_
        self._injctrl = injctrl
        self.loop_state = db_["BiasFBLoopState-Sel"]["value"]
        self.already_set = False

        self.min_bias_voltage = db_["BiasFBMinVoltage-SP"]["value"]  # [V]
        self.max_bias_voltage = db_["BiasFBMaxVoltage-SP"]["value"]  # [V]

        self.model_type = db_["BiasFBModelType-Sel"]["value"]
        self.model_max_num_points = db_["BiasFBModelMaxNrPts-SP"]["value"]
        self.model_auto_fit_rate = db_["BiasFBModelAutoFitEveryNrPts-SP"][
            "value"
        ]
        self.model_auto_fit = db_["BiasFBModelAutoFitParams-Sel"]["value"]
        self.model_update_data = db_["BiasFBModelUpdateData-Sel"]["value"]

        # [V/mA]:
        self.linmodel_angcoeff = db_["BiasFBLinModAngCoeff-SP"]["value"]
        self.linmodel_offcoeff = db_["BiasFBLinModOffCoeff-SP"]["value"]  # [V]

        self._npts_after_fit = 0

        self.bias_data = _np.array(
            db_["BiasFBModelDataBias-Mon"]["value"], dtype=float
        )
        self.injc_data = _np.array(
            db_["BiasFBModelDataInjCurr-Mon"]["value"], dtype=float
        )
        self.gpmodel = None
        self._initialize_models()

        self.do_update_models = False
        pv_ = self._injctrl.currinfo_dev.pv_object("InjCurr-Mon")
        pv_.auto_monitor = True
        pv_.add_callback(self._callback_to_thread)

        # pvname to write method map
        self.map_pv2write = {
            "LoopState-Sel": self.set_loop_state,
            "MinVoltage-SP": self.set_min_voltage,
            "MaxVoltage-SP": self.set_max_voltage,
            "ModelType-Sel": self.set_model_type,
            "ModelMaxNrPts-SP": self.set_max_num_pts,
            "ModelFitParamsNow-Cmd": self.cmd_fit_params,
            "ModelAutoFitParams-Sel": self.set_auto_fit,
            "ModelAutoFitEveryNrPts-SP": self.set_auto_fit_rate,
            "ModelUpdateData-Sel": self.set_update_data,
            "ModelDataBias-SP": self.set_data_bias,
            "ModelDataInjCurr-SP": self.set_data_injcurr,
            "LinModAngCoeff-SP": self.set_lin_ang_coeff,
            "LinModOffCoeff-SP": self.set_lin_off_coeff,
            "GPModNoiseStd-SP": self.set_gp_noise_std,
            "GPModKernStd-SP": self.set_gp_kern_std,
            "GPModKernLenScl-SP": self.set_gp_kern_leng,
            "GPModNoiseStdFit-Sel": self.set_gp_noise_std_fit,
            "GPModKernStdFit-Sel": self.set_gp_kern_std_fit,
            "GPModKernLenSclFit-Sel": self.set_gp_kern_leng_fit,
        }

    def init_database(self):
        """Set initial PV values."""
        for key, pvdbase in self.database.items():
            pvname = key.split(_Const.BIASFB_PROPTY_PREFIX)[1]
            self.run_callbacks(pvname, pvdbase["value"])

    @property
    def use_gpmodel(self):
        """."""
        return self.model_type == _Const.BiasFBModelTypes.GaussianProcess

    @property
    def linmodel_coeffs(self):
        """."""
        ang = self.linmodel_angcoeff
        off = self.linmodel_offcoeff
        return (off, ang)

    @property
    def linmodel_coeffs_inverse(self):
        """."""
        ang = self.linmodel_angcoeff
        off = self.linmodel_offcoeff
        return (-off / ang, 1 / ang)

    def get_delta_current_per_pulse(
        self, per=1, nrpul=1, curr_avg=100, curr_now=99.5, ltime=17 * 3600
    ):
        """."""
        ltime = max(_Const.BIASFB_MINIMUM_LIFETIME, ltime)
        curr_tar = curr_avg / (1 - per / 2 / ltime)
        dcurr = (curr_tar - curr_now) / nrpul
        return dcurr

    def get_bias_voltage(self, dcurr):
        """."""
        dcurr = max(0, dcurr)
        if self.use_gpmodel:
            return self._get_bias_voltage_gpmodel(dcurr)
        return self._get_bias_voltage_linear_model(dcurr)

    def run_callbacks(self, name, *args, **kwd):
        """."""
        name = _Const.BIASFB_PROPTY_PREFIX + name
        if self._injctrl.has_callbacks:
            self._injctrl.run_callbacks(name, *args, **kwd)
            return
        self.database[name]["value"] = args[0]

    def update_log(self, msg):
        """."""
        self._injctrl.run_callbacks("Log-Mon", msg)

    # ###################### Setter methods for IOC ######################
    def set_loop_state(self, value):
        """."""
        self.loop_state = bool(value)
        self.run_callbacks("LoopState-Sts", value)
        return True

    def set_min_voltage(self, value):
        """."""
        self.min_bias_voltage = value
        self._update_models()
        self.run_callbacks("MinVoltage-RB", value)
        return True

    def set_max_voltage(self, value):
        """."""
        self.max_bias_voltage = value
        self._update_models()
        self.run_callbacks("MaxVoltage-RB", value)
        return True

    def set_model_type(self, value):
        """."""
        self.model_type = value
        self.run_callbacks("ModelType-Sts", value)
        return True

    def set_max_num_pts(self, value):
        """."""
        self.model_max_num_points = int(value)
        self._update_models()
        self.run_callbacks("ModelMaxNrPts-RB", value)
        return True

    def cmd_fit_params(self, value):
        """."""
        _ = value
        self._update_models(force_optimize=True)
        return True

    def set_auto_fit(self, value):
        """."""
        self.model_auto_fit = bool(value)
        self.run_callbacks("ModelAutoFitParams-Sts", value)
        return True

    def set_auto_fit_rate(self, value):
        """."""
        self.model_auto_fit_rate = int(value)
        self.run_callbacks("ModelAutoFitEveryNrPts-RB", value)
        return True

    def set_update_data(self, value):
        """."""
        self.model_update_data = bool(value)
        self.run_callbacks("ModelUpdateData-Sts", value)
        return True

    def set_data_bias(self, value):
        """."""
        self.bias_data = _np.array(value)
        max_ = _Const.BIASFB_MAX_DATA_SIZE
        if self.bias_data.size > max_:
            msg = f"WARN: Bias data is too big (>{max_:d}). "
            msg += "Trimming first points."
            _log.warning(msg)
            self.update_log(msg)
            self.bias_data = self.bias_data[-max_:]
        self._update_models()
        self.run_callbacks("ModelDataBias-RB", value)
        return True

    def set_data_injcurr(self, value):
        """."""
        self.injc_data = _np.array(value)
        max_ = _Const.BIASFB_MAX_DATA_SIZE
        if self.injc_data.size > max_:
            msg = f"WARN: InjCurr data is too big (>{max_:d}). "
            msg += "Trimming first points."
            _log.warning(msg)
            self.update_log(msg)
            self.injc_data = self.injc_data[-max_:]
        self._update_models()
        self.run_callbacks("ModelDataInjCurr-RB", value)
        return True

    def set_lin_ang_coeff(self, value):
        """."""
        self.linmodel_angcoeff = value
        self._update_models()
        self.run_callbacks("LinModAngCoeff-RB", value)
        return True

    def set_lin_off_coeff(self, value):
        """."""
        self.linmodel_offcoeff = value
        self._update_models()
        self.run_callbacks("LinModOffCoeff-RB", value)
        return True

    def set_gp_noise_std(self, value):
        """."""
        self.gpmodel.likelihood.variance = value**2
        self._update_models()
        self.run_callbacks("GPModNoiseStd-RB", value)
        return True

    def set_gp_kern_std(self, value):
        """."""
        self.gpmodel.kern.variance = value**2
        self._update_models()
        self.run_callbacks("GPModKernStd-RB", value)
        return True

    def set_gp_kern_leng(self, value):
        """."""
        self.gpmodel.kern.lengthscale = value
        self._update_models()
        self.run_callbacks("GPModKernLenScl-RB", value)
        return True

    def set_gp_noise_std_fit(self, value):
        """."""
        if bool(value):
            self.gpmodel.likelihood.variance.unfix()
        else:
            self.gpmodel.likelihood.variance.fix()
        self.run_callbacks("GPModNoiseStdFit-Sts", value)
        return True

    def set_gp_kern_std_fit(self, value):
        """."""
        if bool(value):
            self.gpmodel.kern.variance.unfix()
        else:
            self.gpmodel.kern.variance.fix()
        self.run_callbacks("GPModKernStdFit-Sts", value)
        return True

    def set_gp_kern_leng_fit(self, value):
        """."""
        if bool(value):
            self.gpmodel.kern.lengthscale.unfix()
        else:
            self.gpmodel.kern.lengthscale.fix()
        self.run_callbacks("GPModKernLenSclFit-Sts", value)
        return True

    # ############ Auxiliary Methods ############
    def _callback_to_thread(self, **kwgs):
        if not self.do_update_models or not self.model_update_data:
            return
        _Thread(target=self._update_data, kwargs=kwgs, daemon=True).start()

    def _initialize_models(self):
        """."""
        self.bias_data = _np.linspace(
            self.min_bias_voltage,
            self.max_bias_voltage,
            self.model_max_num_points,
        )

        self.injc_data = _np_poly.polyval(
            self.bias_data, self.linmodel_coeffs_inverse
        )

        x = self.bias_data[:, None].copy()
        y = self.injc_data[:, None].copy()

        kernel = gpy.kern.RBF(input_dim=1)
        db_ = self.database["BiasFBGPModKernStd-RB"]
        kernel.variance.constrain_bounded(db_["lolim"] ** 2, db_["hilim"] ** 2)
        kernel.variance = db_["value"] ** 2
        if bool(self.database["BiasFBGPModKernStdFit-Sts"]["value"]):
            kernel.variance.unfix()
        else:
            kernel.variance.fix()

        db_ = self.database["BiasFBGPModKernLenScl-RB"]
        kernel.lengthscale.constrain_bounded(db_["lolim"], db_["hilim"])
        kernel.lengthscale = db_["value"]
        if bool(self.database["BiasFBGPModKernLenSclFit-Sts"]["value"]):
            kernel.lengthscale.unfix()
        else:
            kernel.lengthscale.fix()

        gpmodel = gpy.models.GPRegression(x, y, kernel)
        db_ = self.database["BiasFBGPModNoiseStd-RB"]
        gpmodel.likelihood.variance.constrain_bounded(
            db_["lolim"] ** 2, db_["hilim"] ** 2
        )
        gpmodel.likelihood.variance = db_["value"] ** 2
        if bool(self.database["BiasFBGPModNoiseStdFit-Sts"]["value"]):
            gpmodel.likelihood.variance.unfix()
        else:
            gpmodel.likelihood.variance.fix()

        self.gpmodel = gpmodel
        self._update_predictions()

    def _update_data(self, **kwgs):
        bias = self._injctrl.egun_dev.bias.voltage
        dcurr = kwgs["value"]
        if None in {bias, dcurr}:
            return

        # Do not overload data with repeated points:
        xun, cnts = _np.unique(self.bias_data, return_counts=True)
        if bias in xun:
            idx = (xun == bias).nonzero()[0][0]
            if cnts[idx] >= max(2, self.bias_data.size // 5):
                msg = "WARN: Too many data with this abscissa. "
                msg += "Discarding point."
                _log.warning(msg)
                self.update_log(msg)
                return
        self._npts_after_fit += 1

        # Update models data
        self.bias_data = _np.r_[self.bias_data, bias]
        self.injc_data = _np.r_[self.injc_data, dcurr]
        self.bias_data = self.bias_data[-_Const.BIASFB_MAX_DATA_SIZE :]
        self.injc_data = self.injc_data[-_Const.BIASFB_MAX_DATA_SIZE :]
        self._update_models()

    def _update_models(self, force_optimize=False):
        x = _np.r_[self.bias_data, self.linmodel_offcoeff]
        y = _np.r_[self.injc_data, 0]
        x = x[-self.model_max_num_points :]
        y = y[-self.model_max_num_points :]
        if x.size != y.size:
            msg = "WARN: Arrays with incompatible sizes. "
            msg += "Trimming first points of "
            msg += "bias." if x.size > y.size else "injcurr."
            _log.warning(msg)
            self.update_log(msg)
            siz = min(x.size, y.size)
            x = x[-siz:]
            y = y[-siz:]

        if x.size < 2:
            msg = "ERR: Too few data points. "
            msg += "Skipping Model update."
            _log.error(msg)
            self.update_log(msg)
            return

        fit_rate = self.model_auto_fit_rate
        do_opt = self.model_auto_fit
        do_opt &= not (self._npts_after_fit % fit_rate)
        do_opt |= force_optimize

        # Optimize Linear Model
        if do_opt and not self.use_gpmodel:
            self.linmodel_angcoeff = _np_poly.polyfit(
                y, x - self.linmodel_offcoeff, deg=[1]
            )[1]
            self._npts_after_fit = 0

        # update Gaussian Process Model data
        x.shape = (x.size, 1)
        y.shape = (y.size, 1)
        self.gpmodel.set_XY(x, y)

        # Optimize Gaussian Process Model
        if do_opt and self.use_gpmodel:
            self.gpmodel.optimize_restarts(num_restarts=2, verbose=False)
            self._npts_after_fit = 0

        self.run_callbacks("ModelNrPtsAfterFit-Mon", self._npts_after_fit)
        self._update_predictions()

    def _update_predictions(self):
        gp_ = self.gpmodel
        self.run_callbacks(
            "GPModNoiseStd-Mon", float(gp_.likelihood.variance) ** 0.5
        )
        self.run_callbacks("GPModKernStd-Mon", float(gp_.kern.variance) ** 0.5)
        self.run_callbacks("GPModKernLenScl-Mon", float(gp_.kern.lengthscale))
        self.run_callbacks("LinModAngCoeff-Mon", self.linmodel_angcoeff)
        self.run_callbacks("LinModOffCoeff-Mon", self.linmodel_offcoeff)

        self.run_callbacks("ModelDataBias-Mon", self.gpmodel.X.ravel())
        self.run_callbacks("ModelDataInjCurr-Mon", self.gpmodel.Y.ravel())
        self.run_callbacks("ModelNrPts-Mon", self.gpmodel.Y.size)

        injcurr = _np.linspace(0, 1, 100)
        bias_lin = self._get_bias_voltage_linear_model(injcurr)
        bias_gp = self._get_bias_voltage_gpmodel(injcurr)
        self.run_callbacks("LinModInferenceInjCurr-Mon", injcurr)
        self.run_callbacks("LinModInferenceBias-Mon", bias_lin)
        self.run_callbacks("GPModInferenceInjCurr-Mon", injcurr)
        self.run_callbacks("GPModInferenceBias-Mon", bias_gp)

        bias = _np.linspace(self.min_bias_voltage, self.max_bias_voltage, 100)
        injc_lin = _np_poly.polyval(bias, self.linmodel_coeffs_inverse)
        injca_gp, injcs_gp = self._gpmodel_predict(bias)
        injcs_gp = _np.sqrt(injcs_gp)
        self.run_callbacks("LinModPredBias-Mon", bias)
        self.run_callbacks("LinModPredInjCurrAvg-Mon", injc_lin)
        self.run_callbacks("GPModPredBias-Mon", bias)
        self.run_callbacks("GPModPredInjCurrAvg-Mon", injca_gp.ravel())
        self.run_callbacks("GPModPredInjCurrStd-Mon", injcs_gp.ravel())

    def _get_bias_voltage_gpmodel(self, dcurr):
        bias = self._gpmodel_infer_newx(_np.array(dcurr, ndmin=1))
        bias = _np.minimum(bias, self.max_bias_voltage)
        bias = _np.maximum(bias, self.min_bias_voltage)
        return bias if bias.size > 1 else bias[0]

    def _get_bias_voltage_linear_model(self, dcurr):
        bias = _np_poly.polyval(dcurr, self.linmodel_coeffs)
        bias = _np.minimum(bias, self.max_bias_voltage)
        bias = _np.maximum(bias, self.min_bias_voltage)
        bias = _np.array([bias]).ravel()
        return bias if bias.size > 1 else bias[0]

    def _gpmodel_infer_newx(self, y):
        """Infer x given y for current GP model.

        The GP model object has its own infer_newX method, but it is slow and
        didn't give good results in my tests. So I decided to implement this
        simpler version, which works well.

        Args:
            y (numpy.ndarray, (N,)): y's for which we want to infer x.

        Returns:
            x: infered x's.

        """
        x = _np.linspace(self.min_bias_voltage, self.max_bias_voltage, 300)
        ys, _ = self._gpmodel_predict(x)
        ys[ys < 0] = 0
        idm = ys[:, 0].argmax()
        idm += 1
        idx = _np.argmin(_np.abs(ys[:idm] - y[None, :]), axis=0)
        return x[idx, 0]

    def _gpmodel_predict(self, x):
        """Get the GP model prediction of the injected current.

        Args:
            x (numpy.ndarray): bias voltage.

        Returns:
            numpy.ndarray: predicted injected current.

        """
        x.shape = (x.size, 1)
        return self.gpmodel.predict(x)



"""Machine schedule module."""

import re as _re
import copy as _copy

import numpy as _np
from matplotlib import pyplot as _plt
from .. import util as _util
from .. import clientweb as _web
from ..clientarch import Time as _Time
from .utils import interp1d_previous as _interp1d_previous


class MacScheduleData:
    """Machine schedule data."""

    _TAG_FORMAT_BEG = r'(\d+)h(\d+)-(\w)-(\d+\.\d)'
    _TAG_FORMAT_END = r'(\d+)h(\d+)-(\w)'

    _mac_schedule_sdata = dict()
    _mac_schedule_ndata_byshift = dict()
    _mac_schedule_ndata_byday = dict()
    _mac_schedule_ndata_inicurr = dict()

    @staticmethod
    def get_mac_schedule_data(year, formating='plain'):
        """Get machine schedule data for year."""
        MacScheduleData._reload_mac_schedule_data(year)
        if formating == 'plain':
            data = MacScheduleData._mac_schedule_sdata[year]
            mac_schedule = _copy.deepcopy(data)
        elif formating == 'numeric_byshift':
            data = MacScheduleData._mac_schedule_ndata_byshift[year]
            mac_schedule = list(zip(*data))
        elif formating == 'numeric_byday':
            data = MacScheduleData._mac_schedule_ndata_byday[year]
            mac_schedule = list(zip(*data))
        else:
            raise NotImplementedError(
                "machine schedule for formating '{}' "
                "is not defined".format(formating))
        return mac_schedule

    @staticmethod
    def get_users_shift_count(begin, end):
        """Get users shift count for a period."""
        begin, end = MacScheduleData._handle_interval_data(begin, end)
        _, tags = MacScheduleData._get_numeric_data_for_interval(
            begin, end, dtype='macsched_byshift')
        return _np.sum(tags) if begin != end else 0

    @staticmethod
    def get_users_shift_day_count(begin, end, count_sundays=True):
        """Get users shift day count for a period."""
        begin, end = MacScheduleData._handle_interval_data(begin, end)
        tims, tags = MacScheduleData._get_numeric_data_for_interval(
            begin, end, dtype='macsched_byday')

        count = 0
        if begin != end:
            for tim, tag in zip(tims, tags):
                weekday = _Time(tim).weekday()
                if (weekday == 6 and count_sundays) or weekday != 6:
                    count += tag
        return count


    @staticmethod
    def is_user_shift_programmed(
            timestamp=None, datetime=None,
            year=None, month=None, day=None, hour=0, minute=0):
        """Return whether a day is a predefined user shift."""
        timestamp, datetime, ret_uni = MacScheduleData._handle_timestamp_data(
            timestamp, datetime, year, month, day, hour, minute)
        times, tags = MacScheduleData._get_numeric_data_for_interval(
            datetime[0], datetime[-1], dtype='macsched_byshift')
        val = _interp1d_previous(times, tags, timestamp)
        return bool(val) if ret_uni else val

    @staticmethod
    def get_initial_current_programmed(
            timestamp=None, datetime=None,
            year=None, month=None, day=None, hour=0, minute=0):
        """Return initial current for shift."""
        timestamp, datetime, ret_uni = MacScheduleData._handle_timestamp_data(
            timestamp, datetime, year, month, day, hour, minute)
        times, currs = MacScheduleData._get_numeric_data_for_interval(
            datetime[0], datetime[-1], dtype='initial_current')
        val = _interp1d_previous(times, currs, timestamp)
        return val[0] if ret_uni else val

    @staticmethod
    def plot_mac_schedule(year):
        """Get machine schedule data for year."""
        MacScheduleData._reload_mac_schedule_data(year)
        times, tags = MacScheduleData.get_mac_schedule_data(
            year, formating='numeric_byshift')
        days_of_year = len(MacScheduleData._mac_schedule_sdata[year])
        new_timestamp = _np.linspace(times[0], times[-1], days_of_year*24*60)
        new_datetimes = [_Time(ts) for ts in new_timestamp]
        new_tags = _interp1d_previous(times, tags, new_timestamp)

        fig = _plt.figure()
        _plt.gca().xaxis.axis_date()
        _plt.plot(new_datetimes, new_tags, '-')
        _plt.title('Machine Schedule - ' + str(year))
        return fig

    # --- private methods ---

    @staticmethod
    def _reload_mac_schedule_data(year):
        if year in MacScheduleData._mac_schedule_sdata:
            return
        if not _web.server_online():
            raise Exception('could not connect to web server')

        try:
            data, _ = _util.read_text_data(_web.mac_schedule_read(year))
        except Exception:
            print('No data provided for year ' + str(year) + '. '
                  'Getting template data.')
            data, _ = _util.read_text_data(_web.mac_schedule_read('template'))

        databyshift = list()
        databyday = list()
        datainicurr = list()
        flag_bit, inicurr = 0, 0.0
        for datum in data:
            if len(datum) < 2:
                raise Exception(
                    'there is a date ({0}) with problem in {1} '
                    'machine schedule'.format(datum, year))

            month, day = int(datum[0]), int(datum[1])
            if len(datum) == 2:
                timestamp = _Time(year, month, day, 0, 0).timestamp()
                databyshift.append((timestamp, flag_bit))
                databyday.append((timestamp, flag_bit))
                datainicurr.append((timestamp, inicurr))
            else:
                timestamp = _Time(year, month, day, 0, 0).timestamp()
                databyday.append((timestamp, 1))
                for tag in datum[2:]:
                    if 'B' in tag:
                        hour, minute, flag, inicurr = _re.findall(
                            MacScheduleData._TAG_FORMAT_BEG, tag)[0]
                        inicurr = float(inicurr)
                    else:
                        hour, minute, flag = _re.findall(
                            MacScheduleData._TAG_FORMAT_END, tag)[0]
                        inicurr = 0.0
                    flag_bit = 0 if flag == 'E' else 1
                    hour, minute = int(hour), int(minute)
                    timestamp = _Time(
                        year, month, day, hour, minute).timestamp()
                    databyshift.append((timestamp, flag_bit))
                    datainicurr.append((timestamp, inicurr))

        MacScheduleData._mac_schedule_sdata[year] = data
        MacScheduleData._mac_schedule_ndata_byshift[year] = databyshift
        MacScheduleData._mac_schedule_ndata_byday[year] = databyday
        MacScheduleData._mac_schedule_ndata_inicurr[year] = datainicurr

    @staticmethod
    def _handle_timestamp_data(
            timestamp=None, datetime=None, year=None,
            month=None, day=None, hour=0, minute=0):
        ret_uni = False
        if timestamp is not None:
            if not isinstance(timestamp, (list, tuple, _np.ndarray)):
                timestamp = [timestamp, ]
                ret_uni = True
            datetime = [_Time(ts) for ts in timestamp]
        elif datetime is not None:
            if not isinstance(datetime, (list, tuple, _np.ndarray)):
                datetime = [datetime, ]
                ret_uni = True
            timestamp = [dt.timestamp() for dt in datetime]
        elif year is not None:
            ret_uni = True
            datetime = [_Time(year, month, day, hour, minute), ]
            timestamp = [dt.timestamp() for dt in datetime]
        else:
            raise Exception(
                'Enter timestamp, datetime or datetime items data.')
        return timestamp, datetime, ret_uni

    @staticmethod
    def _handle_interval_data(begin, end):
        if isinstance(begin, float):
            begin = _Time(begin)
            end = _Time(end)
        elif isinstance(begin, dict):
            begin = _Time(**begin)
            end = _Time(**end)
        return begin, end

    @staticmethod
    def _get_numeric_data_for_interval(begin, end, dtype='macsched_byshift'):
        times, tags = list(), list()
        for y2l in _np.arange(begin.year, end.year+1):
            MacScheduleData._reload_mac_schedule_data(y2l)
            if dtype == 'macsched_byshift':
                data = MacScheduleData._mac_schedule_ndata_byshift[y2l]
            elif dtype == 'macsched_byday':
                data = MacScheduleData._mac_schedule_ndata_byday[y2l]
            elif dtype == 'initial_current':
                data = MacScheduleData._mac_schedule_ndata_inicurr[y2l]
            ytim, ytag = list(zip(*data))
            times.extend(ytim)
            tags.extend(ytag)
        times, tags = _np.array(times), _np.array(tags)
        if begin != end:
            idcs = _np.where(_np.logical_and(
                times >= begin.timestamp(), times <= end.timestamp()))[0]
            if not idcs.size:
                idcs = _np.searchsorted(times, [begin.timestamp(), ])
                idcs = idcs-1 if idcs[0] != 0 else idcs
            elif idcs[0] != 0:
                idcs = _np.r_[idcs[0]-1, idcs]
            return times[idcs], tags[idcs]
        return times, tags



"""Storage Ring Scrapers."""

from .device import Device as _Device
import time as _time


class _ScraperDev(_Device):
    """."""

    class DEVICES:
        """Devices names."""

        H = 'SI-01SA:DI-ScrapH'
        V = 'SI-01SA:DI-ScrapV'
        ALL = (H, V)

    PROPERTIES_DEFAULT = (
        'Home-Cmd', 'ForceComplete-Mon', 'CoordConvErr-Mon',
        'EnblBacklashComp-Sel', 'EnblBacklashComp-Sts'
        )

    def __init__(self, devname, props2init='all'):
        """."""
        # check if device exists
        if devname not in _ScraperDev.DEVICES.ALL:
            raise NotImplementedError(devname)

        super().__init__(devname, props2init=props2init)

    @property
    def is_force_cmd_complete(self):
        """."""
        return self['ForceComplete-Mon']

    @property
    def is_coordinate_convertion_ok(self):
        """."""
        return not self['CoordConvErr-Mon']

    @property
    def is_backlash_compensation_enabled(self):
        """."""
        return self['EnblBacklashComp-Sts']

    def enable_backlash_compensation(self):
        """."""
        self['EnblBacklashComp-Sel'] = 1

    def disable_backlash_compensation(self):
        """."""
        self['EnblBacklashComp-Sel'] = 0

    def cmd_go_max_aperture(self):
        """Go to maximum slit aperture."""
        self['Home-Cmd'] = 1

    def _wait_finish_moving(self, pv_names, timeout=10):
        """."""
        t0 = _time.time()
        ret1 = self._wait(propty=pv_names[0], value=1, timeout=timeout)
        if not ret1:
            return ret1
        timeout -= _time.time() - t0
        ret2 = self._wait(propty=pv_names[1], value=1, timeout=timeout)
        return ret2

    @staticmethod
    def _check_limits(lims, is_negative=False):
        dlim = lims[1] - lims[0]
        if (dlim == 0) or ((dlim < 0) ^ is_negative):
            stg = 'greater' if is_negative else 'smaller'
            raise ValueError(f'Inner limit must be {stg:s} than outer limit.')


class ScraperH(_ScraperDev):
    """."""

    PROPERTIES_DEFAULT = _ScraperDev.PROPERTIES_DEFAULT + (
        'LeftSlitPos-SP', 'LeftSlitPos-RB',
        'RightSlitPos-SP', 'RightSlitPos-RB',
        'RightDoneMov-Mon', 'LeftDoneMov-Mon',
        'LeftSlitInnerLim-SP', 'LeftSlitInnerLim-RB',
        'LeftSlitOuterLim-SP', 'LeftSlitOuterLim-RB',
        'LeftSlitBacklashDist-SP', 'LeftSlitBacklashDist-RB',
        'LeftMotionCtrl-Cte', 'ForceLeftSlitPos-Cmd',
        'RightSlitInnerLim-SP', 'RightSlitInnerLim-RB',
        'RightSlitOuterLim-SP', 'RightSlitOuterLim-RB',
        'RightSlitBacklashDist-SP', 'RightSlitBacklashDist-RB',
        'RightMotionCtrl-Cte', 'ForceRightSlitPos-Cmd',
        )

    def __init__(self, devname=None, props2init='all'):
        """."""
        if devname is None:
            devname = _ScraperDev.DEVICES.H
        super().__init__(devname, props2init=props2init)

    @property
    def left_slit_pos(self):
        """Left slit position [mm]."""
        return self['LeftSlitPos-RB']

    @property
    def right_slit_pos(self):
        """Right slit position [mm]."""
        return self['RightSlitPos-RB']

    @property
    def is_left_slit_moving(self):
        """."""
        return not self['LeftDoneMov-Mon']

    @property
    def is_right_slit_moving(self):
        """."""
        return not self['RightDoneMov-Mon']

    @property
    def left_slit_limits(self):
        """Limits for left slit."""
        return [self['LeftSlitInnerLim-RB'], self['LeftSlitOuterLim-RB']]

    @left_slit_limits.setter
    def left_slit_limits(self, lims):
        """Set left slit limits."""
        _ScraperDev._check_limits(lims, is_negative=False)
        self['LeftSlitInnerLim-SP'] = lims[0]
        self['LeftSlitOuterLim-SP'] = lims[1]

    @property
    def right_slit_limits(self):
        """Limits for right slit."""
        return [self['RightSlitInnerLim-RB'], self['RightSlitOuterLim-RB']]

    @right_slit_limits.setter
    def right_slit_limits(self, lims):
        """Set right slit limits."""
        _ScraperDev._check_limits(lims, is_negative=True)
        self['RightSlitInnerLim-SP'] = lims[0]
        self['RightSlitOuterLim-SP'] = lims[1]

    @property
    def left_slit_control_prefix(self):
        """."""
        return self['LeftMotionCtrl-Cte']

    @property
    def right_slit_control_prefix(self):
        """."""
        return self['RightMotionCtrl-Cte']

    def wait_slits_finish_moving(self, timeout=10):
        """."""
        names = [slit + 'DoneMov-Mon' for slit in ('Left', 'Right')]
        return self._wait_finish_moving(pv_names=names, timeout=timeout)

    def move_left_slit(self, value):
        """Move left slit to given position [mm]."""
        self['LeftSlitPos-SP'] = value

    def move_right_slit(self, value):
        """Move right slit to given position [mm]."""
        self['RightSlitPos-SP'] = value

    def cmd_force_left_slit(self):
        """Force left slit position."""
        self['ForceLeftSlitPos-Cmd'] = 1
        return self._wait(propty='ForceComplete-Mon', value=1, timeout=10)

    def cmd_force_right_slit(self):
        """Force right slit position."""
        self['ForceRightSlitPos-Cmd'] = 1
        return self._wait(propty='ForceComplete-Mon', value=1, timeout=10)


class ScraperV(_ScraperDev):
    """."""

    PROPERTIES_DEFAULT = _ScraperDev.PROPERTIES_DEFAULT + (
        'TopSlitPos-SP', 'TopSlitPos-RB',
        'BottomSlitPos-SP', 'BottomSlitPos-RB',
        'BottomDoneMov-Mon', 'TopDoneMov-Mon',
        'TopSlitInnerLim-SP', 'TopSlitInnerLim-RB',
        'TopSlitOuterLim-SP', 'TopSlitOuterLim-RB',
        'TopMotionCtrl-Cte', 'ForceTopSlitPos-Cmd',
        'BottomSlitInnerLim-SP', 'BottomSlitInnerLim-RB',
        'BottomSlitOuterLim-SP', 'BottomSlitOuterLim-RB',
        'BottomMotionCtrl-Cte', 'ForceBottomSlitPos-Cmd',
        )

    def __init__(self, devname=None, props2init='all'):
        """."""
        if devname is None:
            devname = _ScraperDev.DEVICES.V
        super().__init__(devname, props2init=props2init)

    @property
    def top_slit_pos(self):
        """Top slit position [mm]."""
        return self['TopSlitPos-RB']

    @property
    def bottom_slit_pos(self):
        """Bottom slit position [mm]."""
        return self['BottomSlitPos-RB']

    @property
    def is_top_slit_moving(self):
        """."""
        return not self['TopDoneMov-Mon']

    @property
    def is_bottom_slit_moving(self):
        """."""
        return not self['BottomDoneMov-Mon']

    @property
    def top_slit_limits(self):
        """Set top slit limits."""
        return [self['TopSlitInnerLim-RB'], self['TopSlitOuterLim-RB']]

    @top_slit_limits.setter
    def top_slit_limits(self, lims):
        """Set top slit limits."""
        _ScraperDev._check_limits(lims, is_negative=False)
        self['TopSlitInnerLim-SP'] = lims[0]
        self['TopSlitOuterLim-SP'] = lims[1]

    @property
    def bottom_slit_limits(self):
        """Set bottom slit limits."""
        return [self['BottomSlitInnerLim-RB'], self['BottomSlitOuterLim-RB']]

    @bottom_slit_limits.setter
    def bottom_slit_limits(self, lims):
        """Set bottom slit limits."""
        _ScraperDev._check_limits(lims, is_negative=True)
        self['BottomSlitInnerLim-SP'] = lims[0]
        self['BottomSlitOuterLim-SP'] = lims[1]

    @property
    def bottom_slit_control_prefix(self):
        """."""
        return self['BottomMotionCtrl-Cte']

    @property
    def top_slit_control_prefix(self):
        """."""
        return self['TopMotionCtrl-Cte']

    def wait_slits_finish_moving(self, timeout=10):
        """."""
        names = [slit + 'DoneMov-Mon' for slit in ('Top', 'Bottom')]
        return self._wait_finish_moving(pv_names=names, timeout=timeout)

    def move_top_slit(self, value):
        """Change top slit position [mm]."""
        self['TopSlitPos-SP'] = value

    def move_bottom_slit(self, value):
        """Change bottom slit position [mm]."""
        self['BottomSlitPos-SP'] = value

    def cmd_force_top_slit(self):
        """Force top slit position."""
        self['ForceTopSlitPos-Cmd'] = 1
        return self._wait(propty='ForceComplete-Mon', value=1, timeout=10)

    def cmd_force_bottom_slit(self):
        """Force bottom slit position."""
        self['ForceBottomSlitPos-Cmd'] = 1
        return self._wait(propty='ForceComplete-Mon', value=1, timeout=10)
